#!/usr/bin/env bash

ASM_LOG_FILEPATH="/var/log/akk_sm.log"

# check log file and if above size limit, delete. Current limit is 100MB
LOG_FILE_SIZE=$(wc -c "${ASM_LOG_FILEPATH}" | awk '{print $1}')
if [[ "${LOG_FILE_SIZE}" -ge 104857600 ]]; then 
    rm -rf "${ASM_LOG_FILEPATH}"
fi

#create log file and give apache ownership
touch "$ASM_LOG_FILEPATH"
chmod u+rwx "$ASM_LOG_FILEPATH"
chown apache:root "$ASM_LOG_FILEPATH"


# Direct all script output to the log file
exec 1>>"${ASM_LOG_FILEPATH}"
exec 2>&1
now=$(date +%Y-%m-%d-T%H:%M:%S%z)
echo 
echo =========================================================================
echo Running Akkadian Service Monitor Script at "${now}"
echo =========================================================================
echo

# Install jq if it's not present since the script cant run without it
rpm -qa | grep -qw jq || yum  -y install jq

# create folder path
APP_CONFIG_FILE_LOCATION="/usr/local/sbin/asm/config.json"
VERSION=$(jq ".version" < ${APP_CONFIG_FILE_LOCATION} | tr -d '"')
FOLDER_NUMBER="v${VERSION%${VERSION#?}}" 
ASM_WEB_DIRECTORY="/var/www/html/asm/${FOLDER_NUMBER}" 

OUTPUT_JSON_FILEPATH="${ASM_WEB_DIRECTORY}/services.json"
SERVICES_FILEPATH="$1"
DEFAULT_SERVICES_FILEPATH="/usr/local/sbin/asm/default_services"
SERVICES=$( jq '.'  < /usr/local/sbin/asm/services_file_template.json)


function readServiceFile() {
    # Read services from specified file, or from default file if the services config file is not provided.
    echo Reading services config file
    while IFS= read -r line || [[ "$line" ]]; do
        if [[ -n "$line" ]]; then
            SERVICES=$(echo "$SERVICES" | jq ".services += { $line: { name:\"\",activeState:\"\",subState:\"\" } }")
        fi
    done < "${SERVICES_FILEPATH:-$DEFAULT_SERVICES_FILEPATH}"
}

function readServiceStatuses() {
    echo Reading service statuses
    # Get an array of service names. Which are the JSON keys.
    readarray -t SERVICE_KEYS < <(echo "${SERVICES}" | jq -r '.services | keys | .[]')

    #loop through keys
    for s in "${SERVICE_KEYS[@]}"; do
        # Find status of each service running.
        if systemctl --all --type service | grep -q "$s";then
            # find service description, active state substate and append to JSON
            SERVICE_DESCRIPTION=$(systemctl show -p Description "${s}" | sed 's/Description=//g')
            SERVICES=$(echo "$SERVICES" | jq ".services.${s} += { name: \"${SERVICE_DESCRIPTION}\" }")

            SERVICE_ACTIVESTATE=$(systemctl show -p ActiveState "${s}" | sed 's/ActiveState=//g')
            # echo "${SERVICE_ACTIVESTATE}"
            SERVICES=$(echo "$SERVICES" | jq ".services.${s} += { activeState: \"${SERVICE_ACTIVESTATE}\" }")
            
            SERVICE_SUBSTATE=$(systemctl show -p SubState "${s}" | sed 's/SubState=//g')
            # echo "${SERVICE_SUBSTATE}"
            SERVICES=$(echo "$SERVICES" | jq ".services.${s} += { subState: \"${SERVICE_SUBSTATE}\" }")
        else
            # list service status as not found
            #echo "$s does NOT exist."
            SERVICES=$(echo "$SERVICES" | jq ".services.${s} += { name: \"Service not found\" }")
            SERVICES=$(echo "$SERVICES" | jq ".services.${s} += { activeState: \"Service not found\" }")
            SERVICES=$(echo "$SERVICES" | jq ".services.${s} += { subState: \"Service not found\" }")
        fi
    done
}

function readHardDiskSpace() {
    echo Getting Hard Disk Space
    # add section to json
    SERVICES=$(echo "${SERVICES}" | jq ".+ { drives: {}  } ")
    # get hard disk params
    DU_LIST=$(df | awk '/dev/ {print $1 " " $5}')
    #split line by line
    readarray -t split <<< "${DU_LIST}"
    for line in "${split[@]}"; do 
        # Split the lines to get the file system name, and the space left
        FS_NAME=$(echo "${line}" | head -n1 | awk '{print $1}')
        FS_PCENT=$(echo "${line}" | head -n1 | awk '{print $2}')

        #add them to services json
        SERVICES=$(echo "${SERVICES}" | jq ".drives  += { \"${FS_NAME}\" : \"${FS_PCENT}\"  }")

    done
    
}

# populates the info section of the SERVICES JSON
function populateJSONInfo() {
    echo Populating ASM JSON file
    # get name, version from config.json and time from system
    CONFIG_NAME=$(jq ".name" < ${APP_CONFIG_FILE_LOCATION})
    CONFIG_VERSION=$(jq ".version" < ${APP_CONFIG_FILE_LOCATION})
    CONFIG_TIMEZONE=$(date +"%Z")
    CONFIG_TIME=$(date +"%Y-%m-%dT%H:%M:%S%z")
    # add info to json
    SERVICES=$(echo "${SERVICES}" | jq ".info += { title: ${CONFIG_NAME} } " )
    SERVICES=$(echo "${SERVICES}" | jq ".info += { version: ${CONFIG_VERSION} } " )
    SERVICES=$(echo "${SERVICES}" | jq ".info += { timezone: \"${CONFIG_TIMEZONE}\" } " )
    SERVICES=$(echo "${SERVICES}" | jq ".info += { time: \"${CONFIG_TIME}\" } " )
}

# mock a log in to PME to test login functionality

function mockPMELogin() {
    echo "Testing APM login function"
    response=$(curl 'https://127.0.0.1/pme/index.php/user/login' \
        -H 'Connection: keep-alive' \
        -H 'Pragma: no-cache' \
        -H 'Cache-Control: no-cache' \
        -H 'Origin: https://127.0.0.1' \
        -H 'Upgrade-Insecure-Requests: 1' \
        -H 'DNT: 1' \
        -H 'Content-Type: application/x-www-form-urlencoded' \
        -H 'Accept: text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9' \
        -H 'Sec-Fetch-Site: same-origin' \
        -H 'Sec-Fetch-Mode: navigate' \
        -H 'Sec-Fetch-User: ?1' \
        -H 'Sec-Fetch-Dest: document' \
        -H 'Referer: https://127.0.0.1/pme/index.php/user/login' \
        -H 'Accept-Language: en-GB,en-US;q=0.9,en;q=0.8' \
        --data 'username=PMEAdmin&password=PMEAdmin&remember=' \
        --compressed \
        --insecure \
        --silent \
        --location \
        -o /dev/null \
        --write-out '%{http_code}')

    if [[ "${response}" == 200  ]]; then 
        echo "active"
    else 
        echo "inactive"
    fi
}

function populateHAInfo() {
    echo Populating HA information
    IS_CLI_PRESENT=$(/usr/local/sbin/asm/is_cli_present.sh)

    if [[ "${IS_CLI_PRESENT}" == false ]]; then 
        # mark ha sections as unavailable and exit
        SERVICES=$(echo "${SERVICES}" | jq ".application.ha.vip += {activeState: \"Akkadian CLI not found on system\"} " )
        SERVICES=$(echo "${SERVICES}" | jq ".application.ha.replicationDB += {activeState: \"Akkadian CLI not found on system\"} " )
        return 0
    fi
    # make calls to various HA functions and return data. this is designed to work with CLI 4 only.
    # Most calls for values have alternate sections that call the code directly if the CLI files aren't found. This ensures functionality on systems that might not have the right CLI scripts.
    local BASE_PATH="/home/akkadianuser/scripts"
    local VIP_STATE
    local VIP_MASTER
    local VIP_ONLINE_NODES
    local VIP_OFFLINE_NODES
    local REPLICATION_STATE
    local REPLICATION_MASTER
    local REPLICATION_ONLINE_NODES
    local REPLICATION_OFFLINE_NODES 

    # VIP STATE
    VIP_STATE=$("sudo ${BASE_PATH}"/high_availability/are_pcsd_pacemaker_and_corosync_running.sh)

    # VIP MASTER
    VIP_MASTER=$("sudo ${BASE_PATH}"/get_master.sh)

    # VIP ONLINE NODES
    VIP_ONLINE_NODES=$("sudo ${BASE_PATH}"/get_onlines.sh)

    # VIP OFFLINE NODES 
    VIP_OFFLINE_NODES=$("sudo ${BASE_PATH}"/get_offlines.sh)

    ###############################################################
    
    # REPLICATION STATE
    REPLICATION_STATE=$("sudo ${BASE_PATH}"/replication/mariadb/is_redundancy_enable_mariadb.sh)
    
    # REPLICATION MASTER
    REPLICATION_MASTER="none"
    
    # REPLICATION NODES
    N=$("sudo ${BASE_PATH}"/ha/get_node_list.sh)
    REP_NODES=($N)
    
    # Process nodes
    # loop through node list and find those online and offline
    for val in "${REP_NODES[@]}"; do
        status=$("sudo ${BASE_PATH}"/check_ip_exists.sh "${val}")
        
        if [[ "${status}" == 1 ]]; then
            #is online
            REPLICATION_ONLINE_NODES+=("${val}")
        else
            REPLICATION_OFFLINE_NODES+=("${val}")
        fi
    done

    # Write info to json
    SERVICES=$(echo "${SERVICES}" | jq ".application.ha.vip += {activeState: \"${VIP_STATE}\"} " )
    SERVICES=$(echo "${SERVICES}" | jq ".application.ha.vip += {masterNode: \"${VIP_MASTER}\"} " )

    # reformat node values to fit in json array
    VIP_ONLINE_NODES=$(echo ${VIP_ONLINE_NODES} | sed 's/[][]//g' ) # remove brackets
    VIP_ONLINE_NODES=("${VIP_ONLINE_NODES}") # convert to bash array
    VIP_OFFLINE_NODES=$(echo ${VIP_OFFLINE_NODES} | sed 's/[][]//g' ) #remove brackets
    VIP_OFFLINE_NODES=("${VIP_OFFLINE_NODES}") # convert to bash array

    for val in "${VIP_ONLINE_NODES[@]}"; do
        SERVICES=$(echo "${SERVICES}" | jq ".application.ha.vip.onlineNodes += [\"${val}\"] " )
    done
    for val in "${VIP_OFFLINE_NODES[@]}"; do
        SERVICES=$(echo "${SERVICES}" | jq ".application.ha.vip.offlineNodes += [\"${val}\"] " )
    done


    # replication values
    SERVICES=$(echo "${SERVICES}" | jq ".application.ha.replicationDB += {activeState: \"${REPLICATION_STATE}\"} " )
    SERVICES=$(echo "${SERVICES}" | jq ".application.ha.replicationDB += {masterNode: \"${REPLICATION_MASTER}\"} " )
    
    
    for val in "${REPLICATION_ONLINE_NODES[@]}"; do
        SERVICES=$(echo "${SERVICES}" | jq ".application.ha.replicationDB.onlineNodes += [\"${val}\"] " )
    done
    for val in "${REPLICATION_OFFLINE_NODES[@]}"; do
        SERVICES=$(echo "${SERVICES}" | jq ".application.ha.replicationDB.offlineNodes += [\"${val}\"] " )
    done


}

# populates the application section of the services json.
function populateApplicationInfo() {
    echo Populating Akkadian Application Information
    local APPLICATION_NAME
    local APPLICATION_VERSION


    # ioncube and php status
    IONCUBE_STATUS=$(php -r "var_dump(extension_loaded('ionCube Loader'));" | grep bool)
    PHP_VERSION=$(php -r 'echo PHP_VERSION;')
    if [[ "${IONCUBE_STATUS}" == "bool(true)" ]]; then 
        # add ioncube status 
        SERVICES=$(echo "${SERVICES}" | jq ".application.php += {ioncube: \"loaded\"} " )
    fi
    if [[ -n "${PHP_VERSION}" && "${PHP_VERSION}" =~ [0-9].[0-9].[0-9][0-9]  ]]; then
        # able to get php version, so must be functional
        SERVICES=$(echo "${SERVICES}" | jq ".application.php += {activeState: \"active\"} " )
    fi

    # Application Server Checks
    if [[ -d /opt/akkadian/aco-server ]]; then
        APPLICATION_NAME="Akkadian Console"

        APPLICATION_VERSION=$(awk -F ';' '{print $1}' <<< grep "." /opt/akkadian/aco-server/wwwroot/version.js | tail -1)

    elif [[  ( -d /var/www/html/pme ) && ( -d /var/www/html/orc ) && ( -d /var/www/html/asb ) ]];then
        APPLICATION_NAME="Akkadian Hosted Collaboration Solutions"
    elif [[ -d /var/www/html/pme ]]; then
        APPLICATION_NAME="Akkadian Provisioning Manager"
        APPLICATION_VERSION=$(grep -w 'version' /var/www/html/pme/application/config/pme.php | awk '{print $3}' | tr -d "'" | tr -d ",")
        
        # function checks
        PME_LOGIN_STATUS=$(mockPMELogin)
        SERVICES=$(echo "${SERVICES}" | jq ".application.functions += {login: \"${PME_LOGIN_STATUS}\"} " )

    elif [[ -d /var/www/html/acm ]]; then

        APPLICATION_NAME="Akkadian Contact Manager"
        
        version_digits_pattern='\d{1,10}\.\d{1,10}\.\d{1,10}(?:\.\d{1,10})?'

        APPLICATION_VERSION=$(grep -P "version.+$version_digits_pattern" /var/www/html/acm/application/config/gd.php | grep -oP "$version_digits_pattern" | awk '{print $1}')

    elif [[ -d /var/www/html/asb ]]; then

        APPLICATION_NAME="Akkadian Site Builder"

        IN=$(git --work-tree=/var/www/html/asb/ --git-dir=/var/www/html/asb/.git describe --tags --exact-match)
        arrIN=("${IN//// }")

        APPLICATION_VERSION=${arrIN[1]}

    elif [[ -f /usr/local/bin/AkkadianCloudConnector ]]; then
        APPLICATION_NAME="Akkadian Cloud Connector"
        APPLICATION_VERSION=$(AkkadianCloudConnector version)
    fi

    # add version and name  to json
    SERVICES=$(echo "${SERVICES}" | jq ".application += {name: \"${APPLICATION_NAME}\"} " )
    SERVICES=$(echo "${SERVICES}" | jq ".application += {version: \"${APPLICATION_VERSION}\"} " )
    
}



function writeServicesJSONFile() {
    echo Writing to the services JSON file
    echo "${SERVICES}" | jq . > "${OUTPUT_JSON_FILEPATH}"
}



# Execution sequence 
readServiceFile
readServiceStatuses
readHardDiskSpace
populateJSONInfo
populateApplicationInfo
populateHAInfo
writeServicesJSONFile

#echo "${SERVICES}"
exit 0